;;; Demostrate Spin Box Template

(require 'xmts)

;;
;; make-spin-box 
;;
;; This function returns a closure that acts upon receiving messages.
;; Messages include: 'up, 'down, and 'set.
;;
;; It automatically gives the spin box a behavior as described in
;; the arguments.
;; 
;; field, up, and down are widgets.
;; min and max are lower and upper limits of an integer.
;; val is the initial value of the spin box.
;; getter is a procedure of 1 integer argument, which must returns
;;   a string for display.
;;
;;
(define (make-spin-box field up down min max val incr time-out getter)

  (define traversing #f)		; Internal variables
  (define direction #f)

					; Generic previous or next func
  (define (prev-or-next check against op incr)
    (lambda ()
      (if (check (op val incr) against)
	  (begin (set! val (op val incr))
		 val)
	  #f)))

  (define prev (prev-or-next >= min - incr))
  (define next (prev-or-next <= max + incr))

  (define (traverse)			; Time out event
    (if traversing 
	(begin (closure direction)
	       (xt:app-add-time-out field time-out traverse))))

  (define (event-handler widget event)	; Handles button press or release
    (let ((type (x:get-event-field event x:button-event:type)))
      (if (= type x:button-press)
	  (begin (set! traversing #t)
;		 (display widget) (newline)
;		 (display event) (newline)
		 (set! direction (if (equal? widget up) 'up 'down))
		 (xmts:process-current-traversal field)
		 (traverse))
	  (set! traversing #f))))

  ; This can be use for slower traversal
  (define (key-event-handler widget event)
    (let ((code (x:get-event-field event x:key-event:keycode)))
      (case code
	((98)  (closure 'up))		; These are server dependent
	((104) (closure 'down)))))
;	(else (format #t "unknown keycode: ~D\n" code)))))

  (define (setter val)			; Set text field with specific value
    (if val
	(xm:text-set-string field (getter val))))

  (define (closure msg . args)		; Object acts upon receiving message
    (case msg
      ((up right next)      (setter (next)))
      ((down left previous) (setter (prev)))
      ((set)                (setter (car args)))
      (else (error "Invalid message sent to spin-box-object" msg))))

  ;; Default time-out value
  (if (not time-out) (set! time-out 100))

  ;; A lot of error checking here.
  (cond ((or (< val min) (> val max))
	 (error 
	 "make-spin-box: initial value must be between min and max values.\n"
	 "min, val, and max are " min val max)
	 #f)
	((not (xmt:is-widget-class field xm:text))
	 (error "make-spin-box: text field widget class not matched." field)
	 #f)
	((not (xmt:is-widget-class up    xm:arrow-button))
	 (error "make-spin-box: up-arrow widget class not matched." up)
	 #f)
	((not (xmt:is-widget-class down  xm:arrow-button))
	 (error "make-spin-box: up-arrow widget class not matched." down)
	 #f)
	((not (integer? time-out))
	 (error "make-spin-box: time-out value must be an integer." time-out)
	 #f)
	(else
	 (xt:set-values field 
			xm:n-cursor-position-visible 
			(xt:get-value field xm:n-editable xt:boolean))
	 (closure 'set val)
	 (xt:add-event-handler up 
			       (+ x:button-press-mask x:button-release-mask)
			       0 event-handler)
	 (xt:add-event-handler down
			       (+ x:button-press-mask x:button-release-mask)
			       0 event-handler)
	 (xt:add-event-handler field x:key-press-mask 0
			       key-event-handler)
	 closure)))

;; Temporary patch for xmt:is-widget-class
(define (xmt:is-widget-class widget class)
  (xt:subclass? widget class))

; Test translations
(define numeric-box #f)
(define (numeric-turn dir) (numeric-box dir))
(define (numeric-focus) (display "focusing") (newline))


;; Main program

(define (main top)

  (let ((w (xmts:create-query-children top '((num . "numeric")
					     (sym . "symbolic")
					     (timer . "timer")
					     (float . "float")
					     (int   . "int")))))

    (define (find-child spinbox name)
      (xmt:name-to-widget spinbox (string-append "." name)))
    
    ; Here is a spin box of values
    (set! numeric-box
	  (make-spin-box (find-child (w 'num) "field") 
			 (find-child (w 'num) "up")
			 (find-child (w 'num) "down")
			 1900 2100 1994 4 ; from 1900 to 2100 for every 5 years
			 #f		  ; use default time-out value
			 number->string))
;			 (lambda (i)
;			   (number->string i))))
  
    ; Here is a spin box of symbol choices
    (let ((choices '(Red Blue Green Yellow Brown Black White)))
      (make-spin-box (find-child (w 'sym) "field") 
		     (find-child (w 'sym) "up")
		     (find-child (w 'sym) "down")
		     0 (- (length choices) 1) 0 1
		     200		; update every 200ms
		     (lambda (i)
		       (symbol->string 
			(list-ref choices i)))))

    ; A timer
    (let ((number->timer (lambda (n)
			   (let ((q (quotient n 60))
				 (r (remainder n 60)))
			     (format #f "   ~2D:~2,'0D" q r)))))
      (make-spin-box (find-child (w 'timer) "field") 
		     (find-child (w 'timer) "up")
		     (find-child (w 'timer) "down")
		     0 300 0 1 30 number->timer))
		     
    ; A floating point spin box
    (make-spin-box (find-child (w 'float) "field") 
		   (find-child (w 'float) "up")
		   (find-child (w 'float) "down")
		   0 100 50 0.125 200 
		   (lambda (val) (format #f "~8,3F" val)))
		     
    ; A formatted integer spin box
    (make-spin-box (find-child (w 'int) "field") 
		   (find-child (w 'int) "up")
		   (find-child (w 'int) "down")
		   0 1000000 1000 100 100 
		   (lambda (val) (format #f "~8:D" val)))
		     
    ))

(xmts:run-app "spin" main)
