;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; xmts.scm (Subroutines for xmtscm)
;;;
;;;   * This is some commonly useful routines.
;;;   * It automatically loads the essential X modules.
;;;   * Most routines are prefixed with 'xmts'.
;;;     If a routine has the same name in Xmt lib, it is prefixed with 'xmt'.
;;;     If a routine has the same name in Motif, it is prefixed with 'xm'.
;;;
;;; Copyright 1995 by Thomas Kwong
;;;
;;; $Id: xmts.scm,v 1.1 1995/08/25 18:27:32 tomkwong Exp tomkwong $
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'x11)
(require 'xt)
(require 'xm)
(require 'xmconst)
(require 'xmtconst)
(require 'xevent)
(require 'xmt)
(require 'xmtdebug)
(require 'assert)
(require 'hash-table)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;            Boilerplates for creating an application
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Upcase the first char, if first one is 'x', upcase the second one also.
(define (xmts:app-name str)
  (let ((new (string-copy str)) (first (char-upcase (string-ref str 0))))
    (string-set! new 0 first)
    (if (char=? first #\X)
	(string-set! new 1 (char-upcase (string-ref new 1))))
    new))

;;; If given one argument, set toplevel to the argument itself.
;;; else just return the toplevel saved.
;(define xmts:toplevel
;  (let ((top #f))
;    (lambda arg
;      (if (null? arg)
;	  top
;	  (set! top (car arg))))))
	  
;;; Init toplevel if not exists, create shell if already exists
(define (xmts:create-toplevel prog)
  (let ((name (xmts:app-name prog)))
    (let ((top (xmt:initialize prog name)))
      ;; re-read program arguments 
      (set! *argv* (program-arguments))
;      (xmts:toplevel top)
      top)))

;;; Loop
(define (xmts:run toplevel proc)
  (xt:realize-widget toplevel)
  (if (not (null? proc)) ((car proc) toplevel))
  (xt:app-main-loop toplevel))

;;; This function creates toplevel, call user-defined procedure, and goes
;;;   into the xt main loop.  So it NEVER RETURNS.
;;; prog should be the name of application.
;;; proc should accept 1 argument (which is the toplevel).
(define (xmts:run-app prog proc . after-realize)
  (let ((top (xmts:create-toplevel prog)))
    (if proc (proc top))
    (xmts:run top after-realize)))

(define (xmts:run-multi-app prog proc . after-realize)
  (let ((top (xmts:create-toplevel prog)))
    (xt:set-values top xm:n-mapped-when-managed #f)
    (if proc (proc top))
    (xmts:run top after-realize)))

(define (xmts:create-app-shell top name)
  (xt:create-popup-shell name xt:application-shell top))

(define (xmts:create-top-level-shell top name)
  (xt:create-popup-shell name xt:top-level-shell top))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                      XMT Convenient Routines
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; These variables and functions are used internally to keep track
;;; of window objects created by xmts:create-query-children.
;;; Set xmts%%maintain-window-objects to #f if you don't want to save
;;; them for some reasons.
(define xmts%%maintain-window-objects #t)

(define xmts%%window-objects '())

(define (xmts%%save-window-object proc)
  (if xmts%%maintain-window-objects
      (set! xmts%%window-objects (cons proc xmts%%window-objects))))

;;; xmts:window-object? 
;;;
;;; User program should use this to check whether a window has
;;; been created by the above xmts:create-query-children procedure.
;;; Returns #t if it is an object created by xmts:create-query-children.
(define (xmts:window-object? obj)
  (if xmts%%maintain-window-objects 
      (if (memq obj xmts%%window-objects) #t #f)
      (procedure? obj)))
    
;;; xmts:create-query-children
;;;
;;; 'info-list' is the data structure containing a list of pairs
;;; of symbols and strings.
;;; 
;;; Returns a closure.
;;;
(define (xmts:create-query-children parent info-list)
  (let ((hash (xmts%%create-query-children-simple parent info-list)))
    (let ((closure (lambda (sym) (xmts:symbol->widget hash sym))))
      (xmts%%save-window-object closure)
      closure)))
      
;;; xmts:create-query-children-simple
;;; 
;;; This one returns only a hash table.
;;;
(define (xmts:create-query-children-simple parent info-list)
  (xmts%%create-query-children-simple parent info-list))

;;;
;;; xmts%%create-query-children-simple 
;;;
;;; Create children and return hash table which contains mappings
;;; from symbol to widget id.
;;;
(define (xmts%%create-query-children-simple parent info-list)

  (define set-wid! (hash-associator eq?))

  ; Find all widget IDs according to the string given
  (define (create-widget-hash)
    (xmt:create-children parent)
    (let loop ((L info-list) (result (make-hash-table 10)))
      (if (null? L)
	  result
	  (let ((info (car L)))
	    (if (and (pair? info) (symbol? (car info)) (string? (cdr info)))
		(let ((wid (xmt:name-to-widget 
			    parent 
			    (string-append "self*" (cdr info)))))
		  (if wid
		      (set-wid! result (car info) wid)
		      (xmts:program-warning 
		       "xmts:create-query-children: Cannot map " 
		       (cdr info) " to widget ID.")))
		(xmts:program-error 
		 "xmts:create-query-children: don't know how to handle: "
		 info))
	    (loop (cdr L) result)))))
  
  ; Double check for validity of argument
  (if (not (list? info-list))
      (begin (xmts:program-warning 
	      "xmts:create-query-children: invalid info-list argument "
	      info-list)
	     #f)
      (create-widget-hash)))


;;; xmts:symbol->widget 
;;;
;;; Find the corresponding widget ID from a symbol inside the hash table.
;;;
(define xmts:symbol->widget (hash-inquirer eq?))

;;; xmts:build-query-dialog-simple 
;;; 
;;; Similar to xmts:create-query-children-simple.  
;;; An extra hash entry will be add with symbol equals 'dialogshell.
(define (xmts:build-query-dialog-simple name parent info-list)
  (let ((shell (xt:create-widget name xm:dialog-shell parent)))
    (let ((hash (xmts:create-query-children-simple shell info-list)))
      (if hash
	  (begin ((hash-associator eq?) hash 'dialogshell shell)
		 hash)
	  #f))))

;;; xmts:build-query-dialog
;;;
;;; The usage of this one is similar to xmts:create-query-children.
;;;
;;; 'name' is the name of the dialog as specified in resource file.
;;;
;;; Returns a closure which maps symbol name(s) into widget IDs.
;;; If the symbol is 'dialogshell, it returns the dialog shell widget ID.
;;; That means the user should NOT use the symbol 'dialogshell for any
;;; widgets under this hierarchy.

(define (xmts:build-query-dialog name parent info-list)
  (let ((hash (xmts:build-query-dialog-simple name parent info-list)))
    (if hash
	(let ((closure (lambda (sym) (xmts:symbol->widget hash sym))))
	  (xmts%%save-window-object closure)
	  closure)
	#f)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Misc ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Bell
(define (xmts:bell w p) (x:bell (xt:display w) p))

;;; List Routines (Under construction...)

; Get number of items
(define (xmts:list-get-size list-widget)
  (xt:get-value list-widget xm:n-item-count xt:integer))

; Set a vector of ascii strings
(define (xmts:list-set-items-ascii list-widget string-vector)
  (xmts:list-set-items list-widget
		       (xmts:vector-map (lambda (s) (xm:string-create s))
					string-vector)))

; Set a vector of xmstring's
(define (xmts:list-set-items list-widget xmstring-vector)
  (xt:set-values list-widget
		 xm:n-items (xm:vector->xmstringtable xmstring-vector)
		 xm:n-item-count (vector-length xmstring-vector)))

; Set a vector of xmstring's for selected items
(define (xmts:list-set-selected-items list-widget xmstring-vector)
  (xt:set-values list-widget
		 xm:n-selected-items (xm:vector->xmstringtable xmstring-vector)
		 xm:n-selected-item-count (vector-length xmstring-vector)))

; Get a vector of ascii string
(define (xmts:list-get-items-ascii list-widget)
  (xmts:vector-map (lambda (xmstr) (xm:string-get-first-segment xmstr))
		   (xmts:list-get-items list-widget)))

; Get a vector of xmstring's 
(define (xmts:list-get-items list-widget)
  (let ((num (xmts:list-get-size list-widget)))
    (xm:xmstringtable->vector (xt:get-value list-widget xm:n-items
					    xt:xmstringtable num))))

; Get the top item position
(define (xmts:list-get-top-item-pos list-widget)
  (xt:get-value list-widget xm:n-top-item-position xt:integer))

; Get the number of visible items
(define (xmts:list-get-visible-item-count list-widget)
  (xt:get-value list-widget xm:n-visible-item-count xt:integer))

; Set the top item position
(define (xmts:list-set-top-item-pos list-widget pos)
  (xt:set-values list-widget xm:n-top-item-position pos))

; Make the specified item visible
(define (xmts:list-make-pos-visible list-widget pos)
  (let ((top (xmts:list-get-top-item-pos list-widget))
	(vis (xmts:list-get-visible-item-count list-widget)))
    (let ((bot (- (+ top vis) 1)))
      (cond ((< pos top) (xmts:list-set-top-item-pos list-widget pos))
	    ((> pos bot) (xmts:list-set-top-item-pos list-widget 
						     (+ (- pos vis) 1)))))))

; Delete all items
(define (xmts:list-delete-all-items list-widget)
  (let ((empty (xm:vector->xmstringtable (make-vector 0))))
    (xt:set-values list-widget 
		   xm:n-items empty
		   xm:n-item-count 0
		   xm:n-selected-items empty
		   xm:n-selected-item-count 0)))

; Delete <number> items starting from <from> inclusive
(define (xmts:list-delete-items list-widget from num)
  (let* ((size  (xmts:list-get-size list-widget))
	 (items (xmts:list-get-items list-widget)))
    (if (< 0 from (+ size 1))
	(let ((new-items (make-vector (- size num)))
	      (from-idx  (- from 1)))
	  (debug items)
	  (debug new-items)
	  (do ((i 0 (+ i 1)))		;copy 'before' items
	      ((= i from-idx))
	    (vector-set! new-items i (vector-ref items i)))
	  (do ((i from-idx (+ i 1)) (j (+ from-idx num) (+ j 1)))
	      ((= j size))
	    (vector-set! new-items i (vector-ref items j)))
	  (xmts:list-set-items list-widget new-items))
	(error "Out of range"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                       Personal Short-Cuts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define xm:n-modify-verify-callback "modifyVerifyCallback")
(define xm:n-motion-verify-callback "motionVerifyCallback")
(define xm:n-extended-selection-callback "extendedSelectionCallback")

;;; Add callbacks
(define (xmts:**add-callback** widget type proc args)
  (if (null? args)
      (xt:add-callback widget type proc)
      (xt:add-callback widget type proc (car args))))

(define (xmts:add-activate widget proc . args)
  (xmts:**add-callback** widget xm:n-activate-callback proc args))

(define (xmts:add-browse widget proc . args)
  (xmts:**add-callback** widget xm:n-browse-selection-callback proc args))

(define (xmts:add-single widget proc . args)
  (xmts:**add-callback** widget xm:n-single-selection-callback proc args))

(define (xmts:add-extended widget proc . args)
  (xmts:**add-callback** widget xm:n-extended-selection-callback proc args))

(define (xmts:add-multiple widget proc . args)
  (xmts:**add-callback** widget xm:n-multiple-selection-callback proc args))

(define (xmts:add-default widget proc . args)
  (xmts:**add-callback** widget xm:n-default-action-callback proc args))

(define (xmts:add-modify widget proc . args)
  (xmts:**add-callback** widget xm:n-modify-verify-callback proc args))

(define (xmts:add-motion widget proc . args)
  (xmts:**add-callback** widget xm:n-motion-verify-callback proc args))

(define (xmts:add-drag widget proc . args)
  (xmts:**add-callback** widget xm:n-drag-callback proc args))

;(define (xmts:add-browse widget proc)
;  (xt:add-callback widget xm:n-browse-selection-callback proc))

;(define (xmts:add-single widget proc)
;  (xt:add-callback widget xm:n-single-selection-callback proc))

;(define (xmts:add-extended widget proc)
;  (xt:add-callback widget xm:n-extended-selection-callback proc))

;(define (xmts:add-default widget proc)
;  (xt:add-callback widget xm:n-default-action-callback proc))

;(define (xmts:add-modify widget proc)
;  (xt:add-callback widget xm:n-modify-verify-callback proc))

;(define (xmts:add-motion widget proc)
;  (xt:add-callback widget xm:n-motion-verify-callback proc))

;;; Set title of a window
(define (xmts:set-title widget string)
  (xt:set-values (xmt:get-shell widget) xm:n-title string))

;;; Process traversal (current)
(define (xmts:process-current-traversal widget)
  (xm:process-traversal widget 0))

;;; Set a label string
(define (xmts:label-set-string label-widget str)
  (xt:set-values label-widget xm:n-label-string (xm:string-create str)))

;;; Getting children of a widget
(define (xmts:get-children widget)
  (if (xt:subclass? widget xt:composite)
      (let ((num (xt:get-value widget xm:n-num-children xt:integer)))
	(if (> num 0)
	    (xt:get-value widget xm:n-children xt:widgetlist num)
	    #f))
      #f))

;;; Recursively set a resource value 
(define (recursive-set-value widget resource value)
  (let ((children (xmts:get-children widget)))
;    (format #t "~A children: ~A\n\n" (xt:name widget) children)
    (if children
	(do ((i 0 (+ i 1)))
	    ((= i (vector-length children)))
	  (recursive-set-value (vector-ref children i) resource value))))
  (xt:set-values widget resource value))
	    
;;; Program Error/Warning Handling
(define (xmts:program-error . args)
  (display "Error: ")
  (let loop ((args args))
    (if (not (null? args))
	(begin (display (car args)) (loop (cdr args)))))
  (newline))

(define (xmts:program-warning . args)
  (display "Warning: ")
  (let loop ((args args))
    (if (not (null? args))
	(begin (display (car args)) (loop (cdr args)))))
  (newline))

;;; How to convert anything to a string (except vector and list).
(define (xmts:obj->string obj)
  (cond ((symbol? obj)  (symbol->string obj))
	((number? obj)  (number->string obj))
	((char? obj)    (make-string 1 obj))
	((boolean? obj) (if obj "#t" "#f"))
	((vector? obj)  (xmts:vector-map xmts:obj->string obj))
	((list? obj)    (map xmts:obj->string obj))
	(else "no string value")))

;;; Map for vector
(define (xmts:vector-map proc vector)
  (let* ((len (vector-length vector))
	 (new (make-vector len)))
    (do ((i 0 (+ i 1)))
	((= i len) new)
      (vector-set! new i (proc (vector-ref vector i))))))

;;; For-each for vector
(define (xmts:vector-for-each proc vector)
  (let* ((len (vector-length vector)))
    (do ((i 0 (+ i 1)))
	((= i len))
      (proc (vector-ref vector i)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                       Resource Management
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (xmts:put-line-resources widget . lines)
  (if (null? lines)
      (error 'xmts:put-line-resources 'missing-arguments)
      (let loop ((L lines))
	(if (not (null? L))
	    (begin (xrm:put-line-resource widget (car L))
		   (loop (cdr L)))))))

(define (xmts:load-resource-file widget file)
  (let ((f (open-file file "r")))
    (if f 
	(do ((line (read-line f) (read-line f)))
	    ((eof-object? line) #t)
	  (xrm:put-line-resource widget line))
	#f)))
	  
